!!!!!!!!!!
module console
  character(*),parameter :: VERSN='4.2.89',TAB=' '//char(9)
  parameter (KON=6,NOK=4)
  logical :: meter=.FALSE.,nolog
  integer :: ltab=0,klok=0
end

!!!!!!!!!!!!!!
module precisn
  include 'precisn.fh'
end

!!!!!!!!!!!!!!
module xplicit
  integer,parameter :: mFACT=20
  integer,parameter,private :: K=selected_int_kind(18)
  integer(K),parameter :: iFACT(0:mFACT)=(/int((/1,1,2,6,24,120,720,5040,40320,362880,3628800,39916800,479001600/),K),6227020800_K &
   ,87178291200_K,1307674368000_K,20922789888000_K,355687428096000_K,6402373705728000_K,121645100408832000_K,2432902008176640000_K/)
  double precision,parameter :: FACT(0:mFACT)=iFACT
contains
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  character(2) function TwoDigits(i)
  intent(in) i
    if (i>9) then; twodigits=char(48+i/10)//char(48+mod(i,10))
             else; twodigits=char(48+i); endif
  end function
  !!!!!!!!!!!!!!!!!!!
  function transpose3(a) result(b)
  include 'precisn.fh'
  intent(in) a; dimension a(:,:,:),b(lbound(a,1):ubound(a,1),lbound(a,3):ubound(a,3),lbound(a,2):ubound(a,2))
    forall (i=lbound(a,1):ubound(a,1)) b(i,:,:)=transpose(a(i,:,:))
  end function
  !!!!!!!!!!!!!!!!!!!!!!!!!
  elemental function sqroot(x) result(r) !square root tolerant of roundoff
  include 'default.fh'
  intent(in) x
    r=ZERO; if (x>ZERO) r=sqrt(x)
  end function
  !!!!!!!!!!!!!!!!!!!!!!!!!!!
  real elemental function to0(x)
  include 'default.fh'
  intent(in) x
    if (abs(x)>BIG) then; to0=sign(BIG,x)
                    else; to0=x; if (abs(to0)<1.E-10) to0=0; endif
  end function
  !!!!!!!!!!!!!!!!!!
  function LowerCase(str) result(res)
  character(*),intent(in) :: str
  character(len(str)) res,tmp
    tmp=str; call ChangeCase(-len(str),tmp); res=tmp
  end function
  !!!!!!!!!!!!!!!!!!
  function UpperCase(str) result(res)
  character(*),intent(in) :: str
  character(len(str)) res,tmp
    tmp=str; call ChangeCase(len(str),tmp); res=tmp
  end function
  !!!!!!!!!!!!!!!!
  function Squeeze(str,chr) result(res)
  character(*),intent(in) :: str,chr
  character(len(str)) res
    res=str
    do
      i=index(trim(res),chr); if (i<1) return
      res(i:)=res(i+1:)
    enddo
  end function
  !!!!!!!!!!!!!!!!!!!!!!!!
  function EvaluateQuadric(pnt,i,cof,grd) result(f)
  include 'default.fh'
  intent(in) pnt,i,cof; intent(out) grd; optional grd
  dimension pnt(3),cof(i:9),grd(3)
    f=ZERO
    if (i/=1) f=dot_product(cof(4:6),pnt*pnt)+cof(7)*pnt(1)*pnt(2)+cof(8)*pnt(1)*pnt(3)+cof(9)*pnt(2)*pnt(3)
    if (i==0) f=f+cof(0)+dot_product(cof(1:3),pnt)
    if (present(grd)) &
     grd=cof(1:3)+TWO*cof(4:6)*pnt+(/cof(7)*pnt(2)+cof(8)*pnt(3),cof(7)*pnt(1)+cof(9)*pnt(3),cof(8)*pnt(1)+cof(9)*pnt(2)/)
  end function
  !!!!!!!!!!!!!!!!!!
  function Inverse34(a) result(b)
  include 'precisn.fh'
  intent(in) a; real(KR),dimension(3,0:3) :: a,b,t
!    t(:,1:)=transpose(a(:,1:))
    t=a; call InvertMatrix(3,t(1,1))
    t(:,0)=-matmul(t(:,1:),a(:,0)); b=t
  end function
  !!!!!!!!!!!!!!!!!
  function Matmul34(a,b) result(c)
  include 'precisn.fh'
  intent(in) a,b
  real(KR),dimension(3,0:3) :: a,b,c,t
    forall (i=1:3) t(i,0)=a(i,0)+dot_product(a(i,1:),b(:,0))
    forall (i=1:3,j=1:3) t(i,j)=dot_product(a(i,1:),b(:,j))
    c=t
  end function
  !!!!!!!!!!!!!!!!!!!!!
  function CrossProduct(a,b) result(c)
  include 'precisn.fh'
  include 'cross.fi'
  end function
  !!!!!!!!!!!!!!!!!
  function MakeList(list,dlim,arg0,arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9)
  character(*),intent(out) :: list
  character(*),intent(in) :: dlim,arg0
  character(*),intent(in),optional :: arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9
    list=adjustl(arg0)
    if (present(arg1)) then; list=trim(list)//dlim//adjustl(arg1)
    if (present(arg2)) then; list=trim(list)//dlim//adjustl(arg2)
    if (present(arg3)) then; list=trim(list)//dlim//adjustl(arg3)
    if (present(arg4)) then; list=trim(list)//dlim//adjustl(arg4)
    if (present(arg5)) then; list=trim(list)//dlim//adjustl(arg5)
    if (present(arg6)) then; list=trim(list)//dlim//adjustl(arg6)
    if (present(arg7)) then; list=trim(list)//dlim//adjustl(arg7)
    if (present(arg8)) then; list=trim(list)//dlim//adjustl(arg8)
    if (present(arg9)) then; list=trim(list)//dlim//adjustl(arg9)
    endif; endif; endif; endif; endif; endif; endif; endif; endif
    MakeList=len_trim(list)
  end function
  !!!!!!!!!!!!!!!!!!!!!
  subroutine Eliminates(string,arg0,arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9)
  character(*),intent(inout) :: string
  character(*),intent(in) :: arg0
  character(*),intent(in),optional :: arg1,arg2,arg3,arg4,arg5,arg6,arg7,arg8,arg9
    call Eliminate(arg0,string)
    if (present(arg1)) then; call Eliminate(arg1,string)
    if (present(arg2)) then; call Eliminate(arg2,string)
    if (present(arg3)) then; call Eliminate(arg3,string)
    if (present(arg4)) then; call Eliminate(arg4,string)
    if (present(arg5)) then; call Eliminate(arg5,string)
    if (present(arg6)) then; call Eliminate(arg6,string)
    if (present(arg7)) then; call Eliminate(arg7,string)
    if (present(arg8)) then; call Eliminate(arg8,string)
    if (present(arg9)) then; call Eliminate(arg9,string)
    endif; endif; endif; endif; endif; endif; endif; endif; endif
  end subroutine
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!
  elemental function Component(z,c) result(f)
  include 'default.fh'
  complex(KR),intent(in) :: z; character(1),intent(in) :: c
    x=real(z); y=aimag(z)
    include 'componen.fi'
  end function
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!
  real pure function Componen2(z,c) result(f)
  include 'default.fh'
  real,intent(in) :: z(2); character(1),intent(in) :: c
    x=z(1); y=z(2)
    include 'componen.fi'
  end function
  !!!!!!!!!!!!!!!!!!!!!!!!!
  elemental function Energy(z) result(f)
  include 'precisn.fh'
  complex(KR),intent(in) :: z
    x=real(z); y=aimag(z); f=x*x+y*y
  end function
end

!!!!!!!!!!!!!!
module options
  use console,only: VERSN
  parameter (mOPT=100,iOPT=3) !3 for old
  character(1) :: notnow='+'
  logical(1) :: chkd(mOPT) !=.FALSE.
  character(iOPT*mOPT) :: opts=' '
  character(200) :: mods(0:mOPT)=' '
  character(*),parameter :: SW='-/+'
  character(*),parameter :: GNP= &
   'reset;unset key;set mxtics;set mytics;set zlab rot by 90;set termopt noenh;set auto fix;set time "GelOE v' &
   //VERSN//'  %Y-%b-%d %H:%M" off 45'
  character(11) :: paw=' '
  integer :: luof=-1,lgnp=len(GNP),kgin=0,ignp=index(GNP,';set time')-1
  character(*),parameter :: DLM=':'
!  character(1) :: DLM=':'
  private iArgCount
  character(18) :: edit=' '
contains
  !!!!!!!!!!!!!!!!!!!!!
  function LocateOption(o,n) result(k)
  character(2),intent(in) :: o
  character(*),intent(in),optional :: n
  character ob*3,nn*2
    k=0; l=len_trim(opts)+2; ob=o !both with at least 1 trailing blank
    nn=notnow; if (iOPT>3.and.present(n)) nn=n
    do j=0,iOPT-3
      i=index(opts(:l),ob,j==0); if (i<1) return
      i=max(1,i-1); if (index(trim(nn),opts(i:i))>0) cycle
      k=(i+3)/iOPT; chkd(k)=.TRUE.; return
    enddo !j
  end function
  !!!!!!!!!!!!!!!!!!!!!
  function NextArgument(i) result(n)
  intent(in) i
  character(260) arg
  logical x
    call OperatingSystem(kos); j=len(SW); if (iOPT<4) j=kos
    do n=i,iArgCount()
      call GetArgument(n,arg); if (index(SW(:j),arg(:1))<1) return !not a switch
      if (kos>1.or.arg(:1)/='/') cycle
      inquire(file=trim(arg),exist=x); if (x) return !unix file spec
    enddo !n
    n=0
  end function
  !!!!!!!!!!!!!!!!!!!!!!!
  subroutine FetchOptions(fil)
  use console,only: KON,NOK
  use xplicit,only: LowerCase
  use precisn,only: KB
  character(*),intent(out) :: fil
  character opt*260 !,q*1
  logical :: off,ex=.TRUE.,all,x
  character(*),parameter :: ED='-i-I/i/I'
    fil=' '; call OperatingSystem(kos)
    if (kos/=2.or.KB/=1) then; lgnp=lgnp-7 !old windows workaround
!$                       else; lgnp=lgnp-7
                         endif
    j=len(SW); if (iOPT<4) j=kos
    if (luof<0) then
      luof=0; n=iargc()
      if (n>=1) then
        do i=n,merge(n,1,iOPT<4),-1 !check just last for old, all for new
          call getarg(i,opt); if (mod(index(ED(:4*min(j,2)),opt(:2)),2)/=1.or.index('nN',opt(3:3))<1) cycle
          edit='edit'; if (kos==2) edit='c:\windows\notepad'; exit
        enddo !i
! look for gin file argument
        do k=1,n
          call getarg(k,opt); if (index(SW(:j),opt(:1))>0) cycle
          if (opt(:1)==',') opt=opt(2:) !strip leading comma
          l=len_trim(opt)
          if (l<5.or.index(opt(:l),'.')<1) then
            opt(l+1:)='.gin'; l=l+4; if (n>2) inquire(file=opt(:l),exist=ex)
            if (edit/=' '.and.ex) call system(trim(edit)//' "'//opt(:l)//'"')
            luof=LogicalUnit(0); open(luof,file=opt(:l),action='read',status='old',iostat=i); if (i/=0) luof=0
          elseif (l>4) then
            if (LowerCase(opt(l-3:l))=='.gin') then
              if (edit/=' ') call system(trim(edit)//' "'//opt(:l)//'"')
              luof=LogicalUnit(0); open(luof,file=opt(:l),action='read',status='old',iostat=i); if (i/=0) luof=0
            endif
          endif
          if (luof<1) cycle
          fil=opt(:l); kgin=k; exit
        enddo !k
      endif
    endif
    i=iOPT-1; l=-i
    do ia=1,iArgCount()
      call GetArgument(ia,opt); if (index(SW(:j),opt(:1))<1) cycle
      if (kos==1.and.opt(:1)=='/') then !possibly unix file spec
        inquire(file=trim(opt),exist=x); if (x) cycle !it is
      endif
      off=LowerCase(opt(2:3))=='no'; if (off) opt(2:)=opt(4:)
      l=l+iOPT; k=scan(trim(opt)//':','=:'); mods((l+3)/iOPT)=opt(k+1:); chkd((l+3)/iOPT)=.FALSE.
      m=max(3,scan(opt(:k),'ABCDEFGHIJKLMNOPQRSTUVWXYZ',.TRUE.)); all=iOPT>3.and.opt(:1)=='/'
      opts(l:l+i)=opt(4-i:2)//opt(m:m); call ChangeCase(i,opts(l:l+i-1))
      do while (off)
        if (all) then; n=index(opts,'-'//opts(l+1:l+i)); if (n<1) n=index(opts,'+'//opts(l+1:l+i))
                 else; n=0; endif
        if (n<1) n=index(opts,opts(l:l+i))
        opts(n:n+i-1)=' '; mods((n+3)/iOPT)=' '; chkd((n+3)/iOPT)=.TRUE.
        if (n==l) then
          l=l-iOPT; exit
        endif
      enddo
    enddo !ia
! defaults for some options
!    if (LenExternal('DLM',q)==1) DLM=q
    call OutputOptions
    ios=1; i=LocateOption('LO')
    if (i>0) then
      opt=mods(i); if (opt==' ') opt='geloe'
      if (index(opt(max(1,len_trim(opt)-3):),'.')<1) opt=trim(opt)//'.log'
      open(KON,file=trim(opt),action='write',status='replace',recl=768) !to hold percent table for up to 90 angles
      if (mods(i)==' ') open(NOK,action='write',iostat=ios) !,status='replace')
    endif
    if (ios/=0) open(NOK,action='write',status='scratch') !suppress expanded log
    if (iOPT<4.or.edit/=' ') return
    if (LocateOption('IN')<1) return
    edit='edit'; if (kos==2) edit='c:\windows\notepad'
  end subroutine
  !!!!!!!!!!!!!!!!!!!!!!!!
  subroutine OutputOptions
  use console,only: ltab,TAB
    i=LocateOption('GP'); paw='-1'; if (LocateOption('IN','-')<1) paw='0'
    if (mods(i)/=' ') then
      paw=mods(i); lgnp=ignp !turn off stamp completely
      i=index(paw,DLM); if (i>0) paw(i:i)=' '
    endif
    i=LocateOption('SD'); ltab=min(i,1); if (mods(i)/=' ') ltab=len(TAB)
  end subroutine
  !!!!!!!!!!!!!!!!!!
  function iArgCount() result(n)
    character(1) c
    n=iargc(); if (luof<1) return
    rewind(luof); n=n-1
    do
      c=' '; read(luof,'(a)',iostat=i) c; if (i/=0.or.c=='=') exit !end of file
      if (c/=' ') n=n+1
    enddo
  end function
  !!!!!!!!!!!!!!!!!!!!!!
  subroutine GetArgument(k,arg)
  intent(in) k; character(*),intent(out) :: arg
  integer i0,i1,i2
    if (luof<1.or.k<kgin) then
      arg=' '; call getarg(k,arg)
    else
      rewind(luof); n=kgin-1
      do
        arg=' '; read(luof,'(a)',iostat=i) arg; if (i/=0.or.arg(:1)=='=') exit !end of file
        if (arg(:1)==' ') cycle !skip indented lines
        n=n+1; if (k/=n) cycle
        i2=index(arg,'"',.TRUE.)+1; i1=index(arg,'"'); i0=index(arg,' ')
        if (i0<i1.or.i2<2) then
          arg(i0:)=' '
        elseif (arg(i2:i2)==' ') then
          arg(i2:)=' '
        endif
        return
      enddo
      arg=' '; call getarg(k-n+kgin,arg)
    endif
  end subroutine
  !!!!!!!!!!!!!!!!!!!!!!!!
  subroutine UnusedOptions
!    write(*,'(1x,a)') trim(opts); write(*,'(20l4)') chkd(1:(len_trim(opts)+1)/iOPT)
    do i=1,len_trim(opts),iOPT
      if (.not.chkd((i+3)/iOPT)) write(*,*) 'Unused or redundent option ',opts(i:i+iOPT-2)
    enddo !i
  end subroutine
  !!!!!!!!!!!!!!!!!!!!!!
  subroutine EchoOptions(syn,n)
  use console,only: KON
  intent(in) n; character(*),intent(in) :: syn(n)
  character(iOPT-1) op; character(1) :: str=SW(2:2)
    l=(len_trim(opts)+1)/iOPT; m=len(op)
    LOOP: do j=1,l
      op=opts(iOPT*j-m:iOPT*j-1)
      do i=j+1,l
        if (opts(iOPT*i-m:iOPT*i-1)/=op) cycle
        chkd(j)=.TRUE.; cycle LOOP !later override
      enddo !i
      if (m>2) str=op(:1)
      do k=1,n
        i=index(syn(k),'|'); if (syn(k)(i+1:i+2)/=op(m-1:m)) cycle
        if (mods(j)/=' ') then; write(KON,*) str//syn(k)(:i-1)//merge(DLM,'=',index(mods(j),'=')>0)//trim(mods(j))
                          else; write(KON,*) str//syn(k)(:i-1); endif
        exit
      enddo !k
      if (k<=n) cycle
      if (mods(j)/=' ') then; write(KON,*) str//op(m-1:m)//'='//trim(mods(j)) !undocumented option
                        else; write(KON,*) str//op(m-1:m); endif
    enddo LOOP !j
  end subroutine
end

!!!!!!!!!!!!!!
module aberrat
  include 'private.fh'
! useful constants
  parameter (ZERO=0,ONE=1,TWO=2,THREE=3,FOUR=4,FIVE=5,SIX=6,SEVEN=7,EIGHT=8,TEN=10,HUN=100 &
   ,HALF=ONE/TWO,THIRD=ONE/THREE,QTR=ONE/FOUR,TENTH=ONE/TEN,HUNTH=ONE/HUN &
   ,DEG=57.29578,PI=3.1415927,BIG=1.D19,TRG=1.D-9,WATER=1.34)
  integer(K4) :: jseed; integer :: kwl
  character(100) :: plt=' '
! maximum number of surfaces
  parameter (mSUR=250,mFB=6,mAD=12/2)
! glass data
  character(24) :: nm(-1:mSUR) !,cat
  character(2) :: dim !could conflict with generic intrinsic of same name
  real(KR) :: wvl(-1:1),rn(0:mSUR),vn(0:mSUR),dn(2,0:mSUR); integer :: ng
  namelist /glass/ nm,rn,vn,dn,ng,dim,wvl !,cat
! first order properties
  real(KR) :: u0,ub0,h0,h1,hb1,uf,efl,fb(mFB),obsr,wv,ubf,om,hbf,ubr,bkf,hap,vig(0:1),scl(0:3),fna,gwin,aim
  integer :: nf,kolr=0,mord,kuf,kdmd
  namelist /first/ u0,ub0,h0,h1,hb1,uf,efl,kolr,nf,fb,obsr,mord,wv,targ,bkf,hbf,kuf,hap,vig,scl,gwin,kdmd,aim
! surface data
  real(KR) :: fl(mSUR),th(0:mSUR),cv(mSUR),rd(mSUR),cc(mSUR),ap(mSUR),ob(mSUR),cp(mSUR),rc(mSUR),as(mSUR),ad(mSUR,2:mAD) &
   ,pp(mSUR),pd(mSUR,2:mAD),pt(mSUR),pc(mSUR),pk(mSUR),sd(mSUR)
  integer :: kth(0:mSUR),kcv(mSUR),kcc(mSUR),kcp(mSUR),kad(mSUR,2:mAD),kmg(mSUR),mg(0:mSUR),ns=0,is0,nad(mSUR),kame(mSUR)=0
  namelist /sdata/ fl,th,cv,rd,cc,mg,ns,ap,ob,cp,rc,as,ad,kth,kcv,kcc,kcp,kad,kmg,pt,pc,pk,pp,pd
! optimization controls
  real(KR) :: vth(0:1,0:mSUR),wf(mFB),dist,glth(2),asth(2),tlen,sdia,bkfd,ubar,bkwd,targ,tolr,delv(2),cos2,cww
  integer :: iter,mult,lseed,klen(2),kdia(2),just,kwasi,mode,merit
  namelist /optim/ cww,mord, &
   wf,dist,glth,asth,tlen,sdia,bkfd,ubar,bkwd,targ,tolr,delv,iter,mult,lseed,klen,kdia,kth,kcv,kcc,kmg,kcp,kad,just,kwasi,mode,cos2
! calculated aberration coefficients
  real(KR) :: w111,w020c(2),w111c(2),w020f,w220f,w420f,w620f,w040f,w311f,w060f,w240f,w242f,w131f &
    ,w040,w131,w222,w220,w311,w060,w151,w422,w420,w511,w333,w331,w242,w240,w040c(2) &
    ,w080,w171,w622,w620,w711,w0x0,pmag,psph,pcma,past,pcrv,pdis,cy3,cy4,cz2,cz3,cz4,posc,violat,wvs(0:2)
  logical :: rstop
  character(*),parameter :: NMM='MIRROR',NM0=' ' !'REFL' 'AIR'
  real(KR) :: pups(0:1,0:1) !pos/siz,en/ex
  real(KR) :: prax(0:6,2,0:mSUR)=0 !n,u,ub,h,hb,c,k: x & y
end
